/* $Id: GaloisMath.c,v 0.1 2004/09/20 20:35:35 jrothwei Exp $ */
/*
** Galois field math routines to support Reed-Solomon error correction.
**
** Copyright (C) 2004  Sensicomm LLC
** Joseph Rothweiler http://www.sensicomm.com
** 
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* Reference: Shu Lin and Daniel J. Costello, Jr., "Error Control Coding:
** Fundamentals and Applications", Prentice-Hall, Englewood Cliffs, NJ, 1983,
** ISBN 0-13-283796-X
*/

/* This version is designed for use on either a workstation or an
** embedded system. So, availability of IO is not assumed unless
** GFSTRINGSTUFF is defined.
*/

/* stdlib is only needed for malloc and free. */
#include <stdlib.h>
#define Free free
#define AllocateArray(typ,siz) ((typ *)malloc(sizeof(typ)*(siz)))
#define ArraySize(x) (sizeof(x)/sizeof(*(x)))

/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/
// #define GFSTRINGSTUFF 1
/* Uncomment one of these 3 lines. */
#define GFSYMBOLSIZE 8
// #define GFSYMBOLSIZE 16
// #define GFSYMBOLSIZE 32

#if GFSYMBOLSIZE==8
#define GFSYMBOL unsigned char
#elif GFSYMBOLSIZE==16
#define GFSYMBOL unsigned short
#elif GFSYMBOLSIZE==32
#define GFSYMBOL unsigned long
#else
#error GFSYMBOLSIZE must be 8, 16, or 32
#endif
typedef struct {
  int  order;       /* Number of bits in a symbol. */
  int  size;        /* Size of the field (2**order). Code length is size-1. */
  GFSYMBOL *atable; /* table[i] = alpha**i. Length size-1. */
  GFSYMBOL *amap;   /* map[alpha**i] = i. Length size-1.   */
  unsigned long minpoly; /* The generator polynomial. */
} GaloisField2
/*Hend;*/;
/*=======1=========2=========3=========4=========5=========6=========7========*/
/* From L&C p. 29. with the X**m term removed. */
static GFSYMBOL GFprimPolys[] = {
/* 0 */         0,         0,         0,       0x3,
/* 4 */      0x03,      0x05,      0x03,      0x09,
/* 8 */     0x01d,
#if GFSYMBOLSIZE > 8
                       0x011,     0x009,     0x005,
/*12 */    0x0053,    0x001b,    0x0443,    0x0003,
/*16 */   0x0100b,
#if GFSYMBOLSIZE > 16
                     0x00009,   0x00081,   0x00027,
/*20 */  0x000009,  0x000005,  0x000003,  0x000021,
/*24 */ 0x0000087,
#endif
#endif
};
static int NumGFprimPolys = ArraySize(GFprimPolys);

/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/ /* Return primitive polynomial for 2**m, or 0 for invalid m. */
GFSYMBOL GaloisPrimPoly(
int m  /* 3 <= m <= NumGFprimPolys */
)
/*Hend;*/ {
	if( (m <= 2)  || (m >= NumGFprimPolys) ) {
		/* Need a better error capture mechanism here. */
		return(0);
	}
	return(GFprimPolys[m]);
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/
/*Usage: GaloisField2 GF; ok=GaloisFieldAllocate(&GF,order); **
** Returns 0 on error. Requires malloc.                      */

int GaloisFieldAllocate(GaloisField2 *GF2,int order)
/*Hend;*/ {
	unsigned long minpoly;
	/* Do this first, so we know order is valid before any malloc's. */
	minpoly = GaloisPrimPoly(order);
	if(minpoly == 0) {
		return 0;
	}

	GF2->order = order;
	GF2->size = 1<<(GF2->order);
	GF2->atable = AllocateArray(GFSYMBOL,2*GF2->size);
	if(GF2->atable == (void *)0) {
		return 0;
	}
	GF2->amap   = GF2->atable + GF2->size;
	GF2->minpoly = minpoly;

	return 1;
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/
/*Usage: GaliosField2 *GF2 = GaloisFieldCreate(order); **
** Returns (void *)0 on error. Requires malloc.        */

GaloisField2 *GaloisFieldCreate(int order)
/*Hend;*/ {
	GaloisField2 *GF2;
	unsigned long minpoly;

	/* Do this first, so we know order is valid before any malloc's. */
	minpoly = GaloisPrimPoly(order);
	if(minpoly == 0) return (void *)0;

	GF2 = AllocateArray(GaloisField2,1);
	if(GF2 == (void *)0) return (void *)0;

	if(GaloisFieldAllocate(GF2,order)==0) return (void *)0;

	return GF2;
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/
/* Call after GaloisFieldAllocate or GaloisFieldCreate.
** Builds tables to convert between Power and Polynomial representations.
** See L&C table 2.8 on p. 33
** Errors: None (but expect a program crash if GF isn't set up yet.)
*/
void GaloisFieldBuild(     /* Table generation (L&C-33) */
  GaloisField2 *GF        /* The computed field parameters and tables. */
)
/*Hend;*/ {
	GFSYMBOL tmp;
	unsigned long CarryMask, CarryFlag;
	unsigned long SymbolMask; /* Only necessary when symbol size < GFSYMBOL */
	int i;

	/* Build the GF table for the nonzero elements:
	 * GF->atable[i] = alpha**i , 0 <=i <= GF->size-1 */

	tmp = GF->atable[0] = 1;   /* alpha**0 */
	CarryMask = GF->size >> 1;
	SymbolMask = GF->size - 1;
	for(i=1; i < GF->size-1; i++) {
		CarryFlag = tmp & CarryMask;
		tmp <<= 1;
		if(CarryFlag) tmp ^=  GF->minpoly;
		tmp &= SymbolMask;
		GF->atable[i] = tmp;
	}

	/* Construct the inverse mapping:
	 * GF->amap[alpha**i] = i */
	for(i=0;i< GF->size-1; i++) GF->amap[GF->atable[i]] = i;
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/
/* Call this when done, only if GaloisFieldAllocate was used. */
void GaloisFieldDeallocate(  /* Deallocate the buffers. */
  GaloisField2 *GF
)
/*Hend;*/ {
	Free((char *)(GF->atable));
	// Free((char *)(GF->amap)); /* Freeing the table frees the map. */

	/* This doesn't really mean anything. */
	GF->order = GF->size = 0;
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/
/* Call this when done, only if GaloisFieldCreate was used. */
void GaloisFieldDestroy(  /* Deallocate the GF buffers and the structure. */
  GaloisField2 *GF
)
/*Hend;*/ {
	GaloisFieldDeallocate(GF);
	Free((char *)GF);
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/ GFSYMBOL GaloisMpy(   /* Return a*b */
  GaloisField2 *GF,
  GFSYMBOL a,
  GFSYMBOL b
)
/*Hend;*/ {
	/* Convert from the n-tuple form to the exponent form,  **
	** add the exponents, and convert back to n-tuple form. */

	if( (a == 0)|| (b == 0) ) return(0);

	return(GF->atable[ ((GF->amap[a] + GF->amap[b]) %(GF->size-1)) ] );
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/ GFSYMBOL GaloisRecip( /* Return 1/b, or 0 if b==0. */
  GaloisField2 *GF,
  GFSYMBOL b
)
/*Hend;*/ {
	GFSYMBOL e_b;



	if( (b == 0) ) {
		/* This is really an error condition. */
		return(0);
	}

	/* First, convert from the n-tuple form to the exponent form. */
	e_b = GF->amap[b];

	if(e_b != 0) e_b = GF->size-1-e_b;  /* Lookup the multiplicative inverse. */

	return(GF->atable[ e_b ] );
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/ GFSYMBOL GaloisDiv( /* Return a/b, or 0 if b==0. */
  GaloisField2 *GF,
  GFSYMBOL a,
  GFSYMBOL b
)
/*Hend;*/ {
	GFSYMBOL e_a, e_b;

	if( (b == 0) ) {
		/* This is an error condition. */
		return(0);
	}
	if( (a == 0) ) return(0); /* This is valid, unless b==0 */

	/* First, convert from the n-tuple form to the exponent form. */

	e_a = GF->amap[a];
	e_b = GF->amap[b];

	if(e_b != 0) e_b = GF->size-1-e_b;  /* Lookup the multiplicative inverse. */

	return(GF->atable[ (e_a + e_b) %(GF->size-1) ] );
}
#ifdef GFSTRINGSTUFF
/* These routines are for debugging only. They **
** convert symbols to human-readable form, in  **
** a format similar to that used in the        **
** examples in L&C.                            */

#include <stdio.h>
#include <unistd.h>
#include <string.h>

/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/ char *GaloisToString( /* Return value as character string. */
  GaloisField2 *GF,
  GFSYMBOL val,    /* The input value in n-tuple form. */
  int fmttype,     /* Input: Output form and optional format flag:
		   ** 0 - Power of alpha: 0 or @n
		   ** 1 - Bitstring (tuple), LSB on right: 0x05 => 00000101
		   ** 2 - Polynomial form. 0x48 -> 0+x3+x6
		   ** | 0x100 : output in hex instead of decimal (for 0 and 2).
		   */
  char *str        /* Returns the tuple as a 0/1 string.
		    * Length GFSYMBOLSIZE+1 including trailing '\0'.*/
)
/*Hend;*/ {
	static char tmpstr[4*GFSYMBOLSIZE+2];
	int type,fmt;

	type=fmttype & 0xff;
	fmt = (fmttype >>8)&0xff;

	/* If str is NULL, use an internal static string. */

	if(str == (void *)0) str = tmpstr;

	str[0] = '\0';
	switch(type) {
		int i;
	case 0:  /* Alpha form. */
		/* Largest supported symbol is 24 bits. So we need     **
		** 9 characters to represent it, + the '@', + the '\0' **
		** giving 11 characters total for the length of str.   */

		if(val == 0) strcpy(str,"0");
		else {
			if(fmt==1) sprintf(str,"@%lx",(unsigned long)(GF->amap[val]));
			else       sprintf(str,"@%lu",(unsigned long)(GF->amap[val]));
		}
		break;

	case 1:  /* bit-string form (LSB to the right). */
		/* Max length is GFSYMBOLSIZE+1. */

		for(i=(GF->size>>1);i>0;i>>=1) {
			if( (val&i)) strcat(str,"1");
			else         strcat(str,"0");
		}
		break;

	case 2:  /* Polynomial. example: 0x48 => 0+x3+x6 */
		/* Max length is less than 4*GFSYMBOLSIZE+2 */

		/* Start with a zero or one. */
		if( (val&1)) strcpy(str,"1");
		else         strcpy(str,"0");

		/* Append the rest of the terms (if any).  */
		for(i=1; i <= GF->order; i++) {
			if( (val&(1<<i))) {
				if(fmt==1) sprintf(tmpstr,"+x%x",i);
				else       sprintf(tmpstr,"+x%d",i);
				strcat(str,tmpstr);
			}
		}
		break;

	default:
		return NULL;
	}
	return str;
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/ char *GaloisPolyToString(  /* Convert polynomial to printable string. */
  GaloisField2 *GF,
  int ord,         /* Order of the polynomial. */
  GFSYMBOL *gpoly, /* Polynomial: Array of length ord+1. */
  char *polstr     /* String representation. Max length approx 10*(order+1).*/
)
/*Hend;*/ {
	char tmpstr[20];
	int i;

	/* Do the first (constant) term. */
	GaloisToString(GF,gpoly[0],0,polstr);

	/* Do the rest of the terms. */
	for(i=1;i<ord+1;i++) {
		strcat(polstr," + ");
		GaloisToString(GF,gpoly[i],0,tmpstr);
		strcat(polstr,tmpstr);
		strcat(polstr,"*");
		sprintf(tmpstr,"X%d",i);
		strcat(polstr,tmpstr);
	}
	return(polstr);
}
#endif
#ifdef TESTPROGRAM
#include <stdio.h>
int main(int argc, char *argv[]) {
	GaloisField2 *GF2;
	GFSYMBOL a, b, r;
	GFSYMBOL ad, bd;
#ifdef GFSTRINGSTUFF
	char bstring[100];
	GFSYMBOL pol[4]={5,4,11,6};
	int i;
#endif

	GF2=GaloisFieldCreate(30);
	if(GF2==NULL) {
		fprintf(stdout,"Trouble creating field of order 30, as expected.\n");
	} else {
		fprintf(stderr,"Oops. This should have failed.\n");
		exit(2);
	}
	GF2=GaloisFieldCreate(8);
	if(GF2==NULL) {
		fprintf(stderr,"Trouble creating field of order 8\n");
	} else {
		fprintf(stdout,"Field of order 8 created.\n");
	}
	GaloisFieldBuild(GF2);

	/* Part of the tables . */

	fprintf(stdout,"Part of the tables.\n");
	fprintf(stdout," %3s %6s %6s | %7s %6s %6s\n","i","a**i","a**i","a**i","i","i");
	fprintf(stdout," %3s %6s %6s | %7s %6s %6s\n","","hex","dec","","hex","dec");
	for(i=0;i<16;i++) {
		fprintf(stdout," %3d %6lx %6ld | %7d",
			i,(unsigned long)(GF2->atable[i]),(unsigned long)(GF2->atable[i]),
			i);
		if(i==0) fprintf(stdout," %6s %6s\n","-","-");
		else     fprintf(stdout," %6lx %6ld\n",
			(unsigned long)(GF2->amap[i]),(unsigned long)(GF2->amap[i]));
	}

	/* Now, some simple tests. */

	fprintf(stdout,"mpy-div test\n");
	a = 2;
	b = 4;
	r = GaloisMpy(GF2,a,b);
	fprintf(stdout," t%d * t%d = t%d\n",a,b,r);
	bd=GaloisDiv(GF2,r,a);
	fprintf(stdout," t%d / t%d = t%d (should be t%d)\n",r,a,bd,b);
	ad=GaloisDiv(GF2,r,b);
	fprintf(stdout," t%d / t%d = t%d (should be t%d)\n",r,b,ad,a);

	fprintf(stdout,"mpy-div test\n");
	a = 3;
	b = 7;
	r = GaloisMpy(GF2,a,b);
	fprintf(stdout," t%d * t%d = t%d\n",a,b,r);
	bd=GaloisDiv(GF2,r,a);
	fprintf(stdout," t%d / t%d = t%d (should be t%d)\n",r,a,bd,b);
	ad=GaloisDiv(GF2,r,b);
	fprintf(stdout," t%d / t%d = t%d (should be t%d)\n",r,b,ad,a);

#ifdef GFSTRINGSTUFF
	GaloisToString(GF2,b,0,bstring);
	fprintf(stdout,"AlphaString t%d --> %s\n",b,bstring);
	GaloisToString(GF2,b,1,bstring);
	fprintf(stdout,"tuple  t%d --> %s\n",b,bstring);
	GaloisToString(GF2,b,2,bstring);
	fprintf(stdout,"XPoly  t%d --> %s\n",b,bstring);
	GaloisPolyToString(GF2,3,pol,bstring);
	fprintf(stdout,"PolyString");
	for(i=0;i<=3;i++) fprintf(stdout," t%d",pol[i]);
	fprintf(stdout," --> %s\n",bstring);
	b=0xff;
	GaloisToString(GF2,b,2,bstring);
	fprintf(stdout,"XPoly  t%d --> %s (%d chars)\n",b,bstring,strlen(bstring));
#endif

	GaloisFieldDestroy(GF2);
	return 0;
}
#endif
