/* $Id: ReedSolomon.c,v 0.1 2004/09/20 20:35:35 jrothwei Exp $ */

/* Copyright (C) 2004  Sensicomm LLC
** Joseph Rothweiler http://www.sensicomm.com
** 
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* Reference: Shu Lin and Daniel J. Costello, Jr., "Error Control Coding:
** Fundamentals and Applications", Prentice-Hall, Englewood Cliffs, NJ, 1983,
** ISBN 0-13-283796-X
*/

// #define DEBUGPRINTOUTS 1

#ifdef DEBUGPRINTOUTS
#include <stdio.h>
// #include <unistd.h>
// #include <stdlib.h>
#endif

#include "GaloisMath.h"

#define max(a,b) ((a)>(b) ? (a) : (b))


/*=======1=========2=========3=========4=========5=========6=========7========*/
/* Construct the generator polynomial. L&C, p. 171 */
/*Hstart*/ void RSGeneratorPoly(
GaloisField2 *GF,
int t,            /* Number of correctable bits. 1 < t < 2**m/2 */
GFSYMBOL *g       /* Length 2*t+1. Returns coefficients in n-tuple **
		  ** (polynomial) form.                  */
)
/*Hend;*/ {
	int i;

	/* The polynomial is
	**          2t
	** g(x) = product [ X+alpha**i]
	**         i=1
	*/

	/* Start by setting the polynomial to (x + alpha) */

	g[0] = GF->atable[1];
	g[1] = 1;
	for(i=2;i<2*t+1;i++) g[i] = 0;

	/* Loop to multiply g by the polynomials (x + a**i) */

	/*    i       k
	** ( sum[g * x ] ) (x + b)
	**   k=0  k
	**
	**   i+1       k     K        k
	** = sum[g   *x ] + sum[g *b*x ])
	**   k=1  k-1       k=0  k
	**
	**            i+1                     k
	** = g *b + ( sum[ (g    + g  * b) * x ] )
	**    0       k=1    k-1    k
	**
	**            i+1         k
	** = g'   + ( sum[ g'  * x ] )
	**    0       k=1   k
	** Note: g_(i+1)==0, so g'_(i+1)=g_i
	*/

	for(i=2; i<=2*t; i++) {
		int k;
		GFSYMBOL b;
		b = GF->atable[i];           /* b = alpha**i */
		k=i;
		g[k] = g[k-1];
		for(k--;k>0;k--) {
			g[k] = g[k-1] ^ GaloisMpy(GF,g[k],b);
		}
		g[0] = GaloisMpy(GF,g[0],b);
	}
	
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/ void RS_encode( 
  GaloisField2 *GF,  /* Input: defines the Galois field. */
  int t,             /* Number of correctable symbols. */
  GFSYMBOL *g,       /* The generator polynomial. */
  GFSYMBOL *indata,  /* Input data symbols. Length GF->size-1-2*t. */
  GFSYMBOL *encdata  /* Output check symbols. Length 2*t. */
)
/*Hend;*/ {
	int i, w;
	int DataLen;
	GFSYMBOL Feedback;

	/* Implement the Encoding structure as shown in L&C, Fig 6.13, p. 172 */

	/* Initialize the check symbols to all zeros. */
	for(i=0;i<2*t;i++) encdata[i] = 0;

	/* Outer loop to clock in the data. */
	DataLen   = GF->size - 1 - 2*t;
	for(w=0;w<DataLen;w++) {
		/* Add the input symbol to the output of the register. */
		Feedback = indata[DataLen-w-1] ^ encdata[2*t-1];
		/* Multiply by the generator polynomial, and add to the */
		/* register. */
		for(i=2*t-1;i>0;i--) {
			GFSYMBOL prod;
			prod=GaloisMpy(GF,Feedback,g[i]);
			encdata[i] = encdata[i-1] ^ prod;
		}
		encdata[0] = GaloisMpy(GF,Feedback,g[0]);
	}
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/* Debug routine to print the berlekamp internals in a format similar to
** the table on p. 175.
*/
#ifdef DEBUGPRINTOUTS
static void berlekampTable(
  FILE *fil,
  GaloisField2 *GF,
  int n,
  GFSYMBOL *d_mu,
  int *l_mu,
  int t,
  int sigmastride,
  GFSYMBOL *t_sig
) {
	int i;
	char dstr[20];
	char polstr[200];

	fprintf(fil," %3s %20s %5s %5s %5s\n","mu","sigma(x)","d_mu","l_mu","mu-l_mu");
	for(i=0;i<n;i++) {
		fprintf(fil,"%3d",i-1);
		GaloisPolyToString(GF,l_mu[i],t_sig+sigmastride*i,polstr);
		fprintf(fil," %20s",polstr);
		GaloisToString(GF,d_mu[i],0,dstr);
		fprintf(fil," %5s",dstr);
		fprintf(fil," %5d",l_mu[i]);
		fprintf(fil," %5d\n",i-1-l_mu[i]);
	}
}
#endif
/*=======1=========2=========3=========4=========5=========6=========7========*/
/* Do I really need to keep all the rows of t_sig around ?? */
static void berlekamp(
  GaloisField2 *GF, int t, GFSYMBOL *S, GFSYMBOL *sigma
) {
	int i, k;
	int mup1;  /* mu + 1 (since arrays can't have a negative index. */
	int sigmastride = 2*t+2;
	GFSYMBOL sigmamatrix[sigmastride*sigmastride];
	GFSYMBOL *sigma_last;
	GFSYMBOL d_mu[2*t+2];
	int l_mu[2*t+2];
	int rhop1, rotp1;
	int best;
	int rho=0;
	GFSYMBOL murho;
	int newl;
	GFSYMBOL *p1, *p2;

	/* Initialize table 6.6, p. 157. */

	for(i=0;i<2*t+2;i++) for(k=0;k<2*t+2;k++) sigmamatrix[sigmastride*i+k] = 0;
	mup1 = 0;  /* mu+1 (row index into table 6.6) */
	d_mu[mup1]     =  1;
	l_mu[mup1]     =  0;
	sigmamatrix[sigmastride*mup1+0] = 1;

	mup1 = 1;
	d_mu[mup1]     =  S[0];
	l_mu[mup1]     =  0;
	sigmamatrix[sigmastride*mup1+0] = 1;

	/* The update loop. mu is the current row, and we are      **
	** computing the entries for row mu+1. The table index for **
	** row mu is mup1.                                         */

	for(mup1 = 1; mup1 <= 2*t; mup1++) {
		GFSYMBOL *sigma_mup1, *sigma_mup2, *sigma_rhop1;
		// printf("mup1 %d\n",mup1); fflush(stdout);
		// printf("sigmamatrix 0x%x sigmastride %d mup1 %d\n",
		// 	(int)sigmamatrix,sigmastride,mup1); fflush(stdout);
		sigma_mup1=sigmamatrix+sigmastride*mup1;
		sigma_mup2=sigmamatrix+sigmastride*(mup1+1);
		/* Step 1, p. 156 */
		if(d_mu[mup1] == 0) {
			/* Copy current sigma to the next row. */
			for(i=0;i<=l_mu[mup1];i++) sigma_mup2[i] = sigma_mup1[i];
			l_mu[mup1+1] = l_mu[mup1];
		} else {
			/* Find the rho row. Step 2, p. 156                 **
			** When 2 rows have the same max l_diff, either can **
			** be used. To match the example in L&C p 175, we   **
			** need to use >best on some compares and >=best    **
			** on others.                                       */
			best = -100;
			for(rotp1 = mup1-1; rotp1 >= 0; rotp1--) {
				if(d_mu[rotp1] != 0) {
					int l_diff;
					l_diff = rotp1-1-l_mu[rotp1];
					if(l_diff > best) {
						best=l_diff;
						rho = rotp1-1;
					}
				}
			}
			rhop1 = rho+1;  /* Table index. */
			sigma_rhop1=sigmamatrix+sigmastride*rhop1;

			/* Compute new sigma per 6.25 */
			murho = GaloisDiv(GF,d_mu[mup1],d_mu[rhop1]);
			for(i=0;i<l_mu[rhop1+1]+1;i++) {
				int t1;
				t1=i+mup1-rhop1;
				// printf("i %d mup1 %d rhop1 %d t1 %d\n",
				//	i,mup1,rhop1,t1); fflush(stdout);
				sigma_mup2[i+mup1-rhop1] =
					GaloisMpy(GF,sigma_rhop1[i],murho);
			}
			newl = max(l_mu[mup1],l_mu[rhop1]+mup1-rhop1);
			for(i=0;i<newl+1;i++) sigma_mup2[i] ^= sigma_mup1[i] ;
			l_mu[mup1+1] = newl;
		}

		/* Compute the new d per 6.27 */

		p1 = S+mup1;
		p2 = sigma_mup2;
		d_mu[mup1+1] = 0;
		for(i=0;i<=l_mu[mup1+1];i++) {
			d_mu[mup1+1] ^= GaloisMpy(GF,*p1--,*p2++);
		}
	}

	/* Copy the result into sigma. */
	sigma_last = sigmamatrix+sigmastride*(2*t+1);
	for(i=0;i<2*t+1;i++) sigma[i] = sigma_last[i];

#ifdef DEBUGPRINTOUTS
	berlekampTable(stdout,GF,2*t,d_mu,l_mu,t,sigmastride,sigmamatrix);
#endif
}
/*=======1=========2=========3=========4=========5=========6=========7========*/
/*Hstart*/
/* Returns the number of error symbols found. If this number     **
** exceeds t, no attempt is made to correct them, and the return **
** value is t+1.                                                 */
int RS_decode(  /* Returns the number of error symbols. */
  GaloisField2 *GF,    /* Must match RS_encode. */
  int t,               /* Must match RS_encode. */
  GFSYMBOL *rcvdata,   /* Input. Length GF->size-1. */
  GFSYMBOL *outdata    /* Output. Can be the same array as rcvdata. */
)
/*Hend;*/ {
	GFSYMBOL Syndrome[2*t+1];  /* Last one is a hack. */
	int s;
	int CodeSize;
	int i;
	int k;
	GFSYMBOL *p1, *p2;
	GFSYMBOL sigma[2*t+2];
	int NumRoots;
	GFSYMBOL val, sum, den;
	GFSYMBOL Z[2*t+2];
	GFSYMBOL beta[2*t+2];
	GFSYMBOL betainv[2*t+2];

	CodeSize = GF->size - 1;

	/* Copy the data to the output (if they are not the same location. */

	if(outdata!=rcvdata) for(i=0;i<CodeSize;i++) outdata[i] = rcvdata[i];
	
	/* First, compute the syndromes, per L&C, Fig. 6.14, p. 174 */

	for (s=0;s<2*t;s++) {
		GFSYMBOL alph;
		GFSYMBOL *ptr;
		GFSYMBOL accum;
		alph = GF->atable[1+s]; /* alph = alpha**(s+1) */
		ptr = rcvdata+CodeSize-1;
		accum = *ptr--;
		for(i=1;i<CodeSize;i++) accum = *ptr-- ^ GaloisMpy(GF,alph,accum);
		Syndrome[s] = accum;

	}
	Syndrome[2*t] = 0;  /* Hack. */
#ifdef DEBUGPRINTOUTS
	{
		char tstrng[256];
		GaloisPolyToString(GF,2*t,Syndrome,tstrng);
		fprintf(stdout,"Syndrome[%d] = %s\n",2*t-1,tstrng);
	}
#endif

	/* Compute the error location polynomials, per L&C, p. 175 */

	berlekamp(GF, t, Syndrome, sigma  ) ;
#ifdef DEBUGPRINTOUTS
	{
		char tstrng[256];
		GaloisPolyToString(GF,2*t,sigma,tstrng);
		fprintf(stdout,"sigma[%d] = %s\n",2*t-1,tstrng);
	}
#endif

	/* Step 3. Find the roots of sigma */

	NumRoots = 0;
	for(k=0;k<GF->size-1;k++) {
		val = GF->atable[k];  /* Step through the field. */
		sum = sigma[2*t];
		for(i=2*t-1;i>=0;i--) {
			sum = sigma[i] ^ GaloisMpy(GF,sum,val);
		}
		if(sum == 0) {
			betainv[NumRoots] = val;
			beta[NumRoots] = GaloisRecip(GF,betainv[NumRoots]);
#ifdef DEBUGPRINTOUTS
			fprintf(stdout,"Found root at k=%d val=%ld 0x%lx beta %ld 0x%lx\n",
				k,(unsigned long)val,(unsigned long)val,
				(unsigned long)(beta[NumRoots]),(unsigned long)(beta[NumRoots]));
#endif
			NumRoots++;

			/* Could skip this test, and count all "errors". */
			if(NumRoots > t)  break;
		}
	}
	if(NumRoots > t) {
		/* Too many errors to correct. */
		return(NumRoots);
	}

	/* Step 4. Form the Z polynomial (6.34). */

	Z[0] = 1;
	for(i=1;i<=t;i++) Z[i] = 0; /* This is just so debug printouts look nice. */

	for(k=1;k<=NumRoots;k++) {
		p1 = Syndrome+k-1;
		p2 = sigma+0;
		Z[k] = 0;
		for(i=0;i<k;i++)  {
			Z[k] ^= GaloisMpy(GF,*p1,*p2);
			p1--;
			p2++;
		}
		Z[k] ^= *p2;
	}
#ifdef DEBUGPRINTOUTS
	{
		char tstrng[256];
		GaloisPolyToString(GF,NumRoots,Z,tstrng);
		fprintf(stdout,"Z[%d] = %s\n",NumRoots,tstrng);
	}
#endif

	/* Step 4b. Substitute the betas, to find the error values. */

	for(k=0;k<NumRoots;k++) {
		GFSYMBOL errsymval;
		GFSYMBOL errloc;
		/* Evalute Z(betainv_k) */
		val = betainv[k];
		sum = Z[NumRoots];
		for(i=NumRoots-1;i>=0;i--) {
			sum = Z[i] ^ GaloisMpy(GF,sum,val);
		}
#ifdef DEBUGPRINTOUTS
		fprintf(stdout,"6.35 numerator %ld 0x%lx\n",
			(unsigned long)sum,(unsigned long)sum);
#endif

		/* Compute the denominator. */

		den = GF->atable[0]; /* Start with den==1. */
		for(i=0;i<NumRoots;i++) if(i != k) {
			GFSYMBOL tmp;
			tmp = 1 ^ GaloisMpy(GF,beta[i],betainv[k]);
			den = GaloisMpy(GF,den,tmp);
		}
#ifdef DEBUGPRINTOUTS
		fprintf(stdout,"6.35 denominator %ld 0x%lx\n",
			(unsigned long)den,(unsigned long)den);
#endif

		errsymval = GaloisDiv(GF,sum,den);

		/* Now, we can correct the error. */

		errloc = GF->amap[beta[k]];
#ifdef DEBUGPRINTOUTS
		fprintf(stdout,"Correcting %3d : loc %3ld err 0x%lx\n",
			k,(unsigned long)errloc,(unsigned long)errsymval);
#endif
		outdata[errloc] ^= errsymval; 
	}

	return(NumRoots);
}
